"Database model error related dictionaries"

from quik_db.utils.formatters import flatten_dict


db_model_errors = {
    "invalid_dialect": """
        Invalid dialect. Only the following dialects are supported:
        mysql, mariadb, mssql, postgresql, oracle, sqlite.
    """,
    "invalid_model_name": """
        Model name provided does not match.
    """,
    "invalid_nesting": """
        Values in a 'MultiDatabaseModel' must be a dictionary. Each
        dictionary represents a database configuration and is validated
        to ensure it matches the schema 'DatabaseModel'.
    """,
    "invalid_params": """
        Connection parameters are missing or incomplete.
    """,
    "invalid_prefix_settings": """
        Schema_prefix is provided, but both prefix_queries and
        prefix_procedures are disabled.
    """,
    "invalid_sqlite_params": """
        Sqlite configuration does not support connection parameters.
        Remove the connection parameters and use_params option if set.
    """,
    "invalid_root": """
        Configuration must be a mapping.
    """,
    "invalid_statevars": """
        State variables are not settable.
    """,
    "invalid_value_type": """
        Configuration must be a mapping.
    """,
    "missing_dialect": """
        Dialect is required for the database configuration.
    """,
    "missing_host": """
        Host is required for the database connection parameters
    """,
    "missing_odbc_driver": """
        Driver is required when using ODBC connections.
    """,
    "missing_params": """
        Connection parameters are required if use_params is set to True.
    """,
    "missing_password": """
        Password is required for the database connection when username
        is provided.
    """,
    "missing_schema": """
        Schema name is required when using the `add_schema_prefix`
        option.
    """,
    "missing_sqlite_url": """
        Sqlite requires a url (path) to the database file. Please
        provide a valid url to the sqlite database file and ensure it
        is in the correct fmt. e.g. (url: '/path/to/sqlite.db')
    """,
    "multiple_default_db": """
        Only one default database can be defined.
    """,
    "invalid_url_and_params": """
        Url and connection_params cannot be provided together.
    """,
    "missing_url_or_params": """
        Either `url` or `connection_params` must be provided for
        non-sqlite dialects.
    """,
}

db_model_errors = flatten_dict(db_model_errors)

db_validation_errors = {
    "invalid_dialect": {
        "loc": ("dialect",),
        "type": "value_error",
        "ctx": {
            "error": db_model_errors["invalid_dialect"],
        },
    },
    "invalid_nesting": {
        "loc": ("__root__",),
        "type": "value_error",
        "ctx": {
            "error": db_model_errors["invalid_nesting"],
        },
    },
    "invalid_value_type": {
        "loc": ("__root__",),
        "type": "value_error",
        "ctx": {
            "error": db_model_errors["invalid_value_type"],
        },
    },
    "invalid_params": {
        "loc": ("connection_params",),
        "type": "missing",
        "ctx": {
            "error": db_model_errors["invalid_params"],
            "example": {
                "connection_params": {
                    "host": "localhost",
                    "username": "user",
                    "password": "pass1",
                    "port": 3306,
                    "database": "mydb",
                },
            },
        },
    },
    "invalid_prefix_settings": {
        "loc": ("query_settings",),
        "type": "value_error",
        "ctx": {
            "error": db_model_errors["invalid_prefix_settings"],
        },
    },
    "invalid_url_and_params": {
        "loc": ("url", "connection_params"),
        "type": "value_error",
        "ctx": {
            "error": db_model_errors["invalid_url_and_params"],
        },
    },
    "missing_dialect": {
        "loc": ("dialect",),
        "type": "missing",
        "ctx": {
            "error": db_model_errors["missing_dialect"],
            "valid_values": [
                "mysql",
                "mariadb",
                "mssql",
                "postgresql",
                "oracle",
                "sqlite",
            ],
        },
    },
    "missing_host": {
        "loc": ("connection_params.host",),
        "type": "missing",
        "ctx": {
            "error": db_model_errors["missing_host"],
        },
    },
    "missing_params": {
        "loc": ("connection_params",),
        "type": "missing",
        "ctx": {"error": db_model_errors["missing_params"]},
    },
    "missing_password": {
        "loc": ("connection_params.password",),
        "type": "missing",
        "ctx": {
            "error": db_model_errors["missing_password"],
        },
    },
    "missing_odbc_driver": {
        "loc": ("driver",),
        "type": "missing",
        "ctx": {
            "error": db_model_errors["missing_odbc_driver"],
        },
    },
    "missing_schema": {
        "loc": ("query_settings.schema_name",),
        "type": "missing",
        "ctx": {
            "error": db_model_errors["missing_schema"],
        },
    },
    "missing_sqlite_url": {
        "loc": ("path",),
        "type": "missing",
        "ctx": {
            "error": db_model_errors["missing_sqlite_url"],
            "example": {"path": "/path/to/sqlite.db"},
        },
    },
    "multiple_default_db": {
        "loc": ("default",),
        "type": "value_error",
        "ctx": {
            "error": db_model_errors["multiple_default_db"],
        },
    },
    "missing_url_or_params": {
        "loc": ("url", "connection_params"),
        "type": "missing",
        "ctx": {
            "error": db_model_errors["missing_url_or_params"],
            "example": {
                "url": "sqlite:///path/to/sqlite.db",
                "connection_params": {
                    "host": "localhost",
                    "username": "user",
                    "password": "pass1",
                    "port": 3306,
                    "database": "mydb",
                },
            },
        },
    },
}
db_validation_errors = flatten_dict(db_validation_errors)



"In development. Class for tracking changes to the connection object"

from __future__ import annotations

from datetime import datetime
from typing import (
    Any,
    List,
    Optional,
    Tuple,
    Type,
)

import wrapt
from pydantic import BaseModel


ChangeRecord = Tuple[
    str, str, str, Optional[dict[str, Any]], Optional[Any], datetime
]


class ChangeTrackingProxy(wrapt.ObjectProxy):
    """
    A proxy to track changes in attributes and methods, including
    nested ones.

    Args:
        wrapped: The object to be wrapped and monitored for changes.
    """

    def __init__(self, wrapped: Any) -> None:
        super().__init__(wrapped)
        self._changes: List[ChangeRecord] = []
        self._self_debug = getattr(wrapped, "debug", False)

    def record_change(
        self,
        parent: str,
        path: str,
        change_type: str,
        params: Optional[dict[str, Any]] = None,
        new_value: Optional[Any] = None,
    ) -> None:
        """
        Record a change with specified details.

        Args:
            parent: The name of the parent object.
            path: The path to the changed attribute or method.
            change_type: The type of change, e.g., 'attribute_change'.
            params: Optional; any parameters involved in the change.
            new_value: Optional; the new value after the change.
        """
        timestamp = datetime.now()
        self._changes.append((
            parent,
            path,
            change_type,
            params,
            new_value,
            timestamp,
        ))

    def __setattr__(self, name: str, value: Any) -> None:
        if self._self_debug and not name.startswith("_"):
            old_value = getattr(self, name, None)
            if old_value != value:
                full_path = f"{self.__wrapped__.__class__.__name__}.{name}"
                self.record_change(
                    parent=self.__wrapped__.__class__.__name__,
                    path=full_path,
                    change_type="attribute_change",
                    new_value=value,
                )
        super().__setattr__(name, value)

    def __getattr__(self, name: str) -> Any:
        attr = super().__getattr__(name)
        if isinstance(attr, (BaseModel, dict, list)) and not isinstance(
            attr, ChangeTrackingProxy
        ):
            return ChangeTrackingProxy(attr)
        elif callable(attr) and not name.startswith("_"):

            @wrapt.decorator
            def track_call(wrapped, instance, args, kwargs):
                if self._self_debug:
                    full_path = f"{self.__wrapped__.__class__.__name__}.{name}"
                    self.record_change(
                        parent=self.__wrapped__.__class__.__name__,
                        path=full_path,
                        change_type="method_call",
                        params={"args": args, "kwargs": kwargs},
                    )
                return wrapped(*args, **kwargs)

            return track_call(attr)
        return attr


def with_tracking(cls: Type[BaseModel]) -> Type[BaseModel]:
    """
    Class factory to apply ChangeTrackingProxy subclassing if
    `debug` is True.

    Args:
        cls: The target class to apply tracking to if its debug
            attribute is True.

    Returns:
        The tracked class if `debug` is True, else the original class.
    """
    if getattr(cls, "debug", False):

        class TrackedDatabaseConnection(ChangeTrackingProxy, cls):
            """
            TrackedDatabaseConnection is a class that extends
            ChangeTrackingProxy.
            """

            def __init__(self, *args, **kwargs):
                super().__init__(cls(*args, **kwargs))

        return TrackedDatabaseConnection
    return cls



"""
This module provides utility functions for working with SQL queries and
database configurations.
"""

from __future__ import annotations

import re
from typing import (
    Any,
    Callable,
    Set,
)

import sqlglot
import sqlparse
from sqlglot import exp
from sqlparse.tokens import DML


def is_stored_procedure(query: str) -> bool:
    """
    Determine if a given SQL query string is a stored procedure call.

    This function checks if the provided SQL query matches common
    patterns used to call stored procedures in various SQL dialects.

    Args:
        query (str): The SQL query string to be checked.

    Returns:
        bool: True if the query is identified as a stored procedure
            call, False otherwise.

    Examples:
        >>> is_stored_procedure("EXEC my_stored_procedure")
        True
        >>> is_stored_procedure("SELECT * FROM my_table")
        False

    """
    query = query.strip().lower()

    # Patterns to detect stored procedure calls
    patterns = [
        r"^exec\s",  # SQL Server, Sybase
        r"^execute\s",  # SQL Server, Sybase (alternative)
        r"^call\s",  # MySQL, PostgreSQL, Oracle
        r"^begin\s",  # PL/SQL block in Oracle
        r"^declare\s",  # PL/SQL anonymous block
    ]

    return any(re.match(pattern, query) for pattern in patterns)


def has_sorting(sql: str) -> bool:
    """
    Check if an SQL query contains an "ORDER BY" clause outside
    of subqueries.

    Args:
        sql (str): The SQL query string to be checked.

    Returns:
        bool: True if the SQL query contains an "ORDER BY" clause outside of
        subqueries, False otherwise.

    """
    sql = re.sub(r"\([^()]*\)", "", sql)

    # Check if "ORDER BY" exists outside of subqueries
    order_by_pattern = re.compile(r"\bORDER\s+BY\b", re.IGNORECASE)

    return bool(order_by_pattern.search(sql))


def has_paging(sql: str) -> bool:
    """
    Detect paging with limits and offsets in an SQL query.

    Args:
        sql (str): The SQL query to analyze.

    Returns:
        bool: True if paging is detected, False otherwise.

    """
    # Regular expressions for common paging patterns with diff dialects
    paging_patterns = [
        # MySQL
        r"\bLIMIT\s+(\d+)\s*(?:OFFSET\s+(\d+))?\b",
        # PostgreSQL
        r"\bLIMIT\s+(\d+)\b(?:\s+OFFSET\s+(\d+)\b)?",
        # SQL Server
        r"\bTOP\s+(\d+)\b(?:\s+OFFSET\s+(\d+)\b)?",
        # Oracle
        r"\bROWNUM\s*\b(?:\s*\bBETWEEN\s*\d+\s*AND\s*\d+\b|\s*\b<=?\s*\d+\b)",
        # SQLite
        r"\bLIMIT\s+(\d+)\s*(?:OFFSET\s+(\d+))?\b",
    ]

    for pattern in paging_patterns:
        match = re.search(pattern, sql, re.IGNORECASE)
        if match:
            return True

    return False


def ms_param_string(p: dict[str, Any]) -> str:
    """
    Return a string of SQL parameters for MSSQL.

    Args:
        p (dict[str, Any]): The parameters to format.

    Returns:
        str: The formatted parameter string.

    """
    return ", ".join([f"@{key}=:{key}" for key in p]) if p else ""


def get_default_config(config: dict) -> bool | str:
    """
    Retrieve the default database configuration from a given
    configuration dictionary.

    Args:
        config (dict): A dictionary containing database configurations.
            Each key is a database name, and each value is a dictionary
            with configuration details, including a "default" key.

    Returns:
        bool | str: The configuration dictionary of the default database
        if found, otherwise None.

    """
    for cfg in config.values():
        if cfg.get("default"):
            return cfg
    return None


def get_default_db_count(config: dict) -> int:
    """
    Calculate the number of default databases in the given configuration.

    Args:
        config (dict): The database configuration dictionary that matches the
        MultiDatabaseModel schema.

    Returns:
        int: The count of databases marked as default in the configuration.

    """
    return sum(
        1
        for db in config.values()
        if isinstance(db, dict) and db.get("default")
    )


def is_dml_query(statement: str) -> bool:
    """
    Detects if the given query is a DML query (NSERT, UPDATE, DELETE).
    """
    parsed = sqlparse.parse(statement)
    for stmt in parsed:
        for token in stmt.tokens:
            if token.ttype == DML:
                return token.value.upper() in {"INSERT", "UPDATE", "DELETE"}
    return False


def build_sql_proc_params(params: dict[str, str]) -> str:
    """Formats parameters for SQL stored procedure query."""
    return f"({', '.join(f':{key}' for key in params)})" if params else ""


def process_params(
    params: dict[str, str] | tuple[str, str] | tuple[tuple[str, str]],
    handlers: list[Callable],
) -> dict[str, str] | tuple[tuple[str, str]]:
    """
    Preprocess the parameters before passing them to the database query.

    Args:
        params (dict[str, Any] | tuple[tuple[str, Any], ...] | None):
            The parameters to be processed. Can be a dictionary or a tuple
            of key-value pairs.
        processors (list[Callable[[Any], Any]] | None):
            A list of processor functions to apply to each value in the
            parameters. If None,no processing will occur, and the original
            parameters will be returned.

    Returns:
        dict[str, Any] | tuple[tuple[str, Any], ...]:
            The processed parameters as a dictionary or the original type
            if no processing occurs.

    """
    if not params or not handlers:
        return params
    if isinstance(params, dict):
        return {
            key: h(value) for key, value in params.items() for h in handlers
        }
    if isinstance(params, tuple):
        return tuple(
            (key, h(value)) for key, value in params for h in handlers
        )


def return_mapped_dialect(dialect: str) -> str:
    """return mapped dialect for sqlglot"""
    map = {
        "mysql": "mysql",
        "mariadb": "mysql",
        "postgresql": "postgres",
        "sqlite": "sqlite",
        "oracle": "oracle",
        "msql": "sqlserver",
    }

    return map.get(dialect, "")


def is_temp_table(table_name: str, dialect: str) -> bool:
    """
    Determine if the given table name is a temporary table based
    on the provided SQL dialect.

    Args:
        table_name (str): The name of the table to check.
        dialect (str): The SQL dialect in use.

    Returns:
        bool: True if the table is a temporary table, False otherwise.
    """
    if dialect == "tsql":
        return table_name.startswith("#")
    # Add conditions for other dialects if necessary
    return False


def apply_schema_prefix(
    node: exp.Expression, schema_prefix: str, cte_names: Set[str], dialect: str
) -> None:
    """
    Recursively apply a schema prefix to table nodes within an SQL AST.

    Args:
        node (exp.Expression): The current node in the SQL AST.
        schema_prefix (str): The schema prefix to apply.
        cte_names (Set[str]): The set of CTE names to exclude from prefixing.
        dialect (str): The SQL dialect in use.
    """
    match node:
        case exp.Table():
            table_name = node.this.name.lower()
            if table_name not in cte_names and not is_temp_table(
                table_name, dialect
            ):
                # Apply schema prefix
                node.set("db", exp.Identifier(this=schema_prefix))

    for child in node.args.values():
        match child:
            case exp.Expression():
                apply_schema_prefix(child, schema_prefix, cte_names, dialect)
            case list():
                for item in child:
                    if isinstance(item, exp.Expression):
                        apply_schema_prefix(
                            item, schema_prefix, cte_names, dialect
                        )


def apply_schema_to_query(
    query: str, schema_prefix: str, dialect: str = "mysql"
) -> str:
    """
    Modify a SQL query by applying a schema prefix to table references.

    Args:
        query (str): The original SQL query to modify.
        schema_prefix (str): The schema prefix to apply.
        dialect (str, optional): The SQL dialect in use. Defaults
            to 'mysql'.

    Returns:
        str: The modified SQL query with schema prefixes applied.
    """
    dialect = return_mapped_dialect(dialect)
    # Parse the SQL query into an AST with the specified dialect
    ast = sqlglot.parse_one(query, read=dialect)

    # Collect the names of CTEs
    cte_names: Set[str] = {
        cte.alias_or_name.lower() for cte in ast.find_all(exp.CTE)
    }

    # Apply schema prefix to table references
    apply_schema_prefix(ast, schema_prefix, cte_names, dialect)

    return ast.sql(dialect=dialect)



"Initialize the model module."
# ruff: noqa: F401, F811, I002
# flake8: noqa: E401
# pylint: disable=unused-import
# pyright: reportUnusedImport=false

from quik_db.model.enums import (
    ConnectionState,
    Dialect,
    ExecutionState,
    QueryType,
)
from quik_db.model.exceptions import ModelValidationError
from quik_db.model.models import DatabaseModel, MultiDatabaseModel
from quik_db.model.validator import model_validator



"""This module contains enums for database operations."""

from enum import Enum, auto


class QueryType(str, Enum):
    """
    Enum for database operation types.

    This enum specifies the types of database operations that can be
    performed.
    """

    PROCEDURE = "procedure"  # Represents a stored procedure operation.
    FUNCTION = "function"  # Represents a function call operation.
    QUERY = "query"  # Represents a standard SQL query operation.


class Dialect(str, Enum):
    """
    Enum for supported database dialects.

    This enum defines the types of database dialects supported by the
    application.
    """

    MYSQL = "mysql"
    MARIADB = "mariadb"
    MSSQL = "mssql"
    POSTGRESQL = "postgresql"
    ORACLE = "oracle"
    SQLITE = "sqlite"


class ExecutionState(str, Enum):
    """
    Enum for database execution states.

    Represents the current state of database execution operations.
    """

    IDLE = "idle"  # Execution is idle.
    ACTIVE = "active"  # Execution is active.
    ERROR = "error"  # Execution encountered an error.


class ConnectionState(Enum):
    """
    Enum representing the state of a database connection.

    Used to track the status of the database connection.
    """

    CONNECTED = auto()  # Connection is active.
    DISCONNECTED = auto()  # Connection is not active.
    ERROR = auto()  # Connection encountered an error.



"""
Database connection classes used to instantiate database connections using
mappings.

"""

# ruff: noqa: D102
# flake8: noqa: E501
# pyright: reportAttributeAccessIssue=false, reportArgumentType=false
from __future__ import annotations

from abc import ABC, abstractmethod
from pprint import pformat
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    override,
)
from warnings import warn

from sqlalchemy import (
    URL,
    create_engine,
    text,
)
from sqlalchemy.orm import (
    Session,
    scoped_session,
    sessionmaker,
)

from quik_db.core.handlers import handler
from quik_db.dictionaries.db_errors import db_errors as errors
from quik_db.model.enums import (
    ConnectionState,
    Dialect,
    ExecutionState,
    QueryType,
)
from quik_db.model.models import (
    DatabaseModel,
    MultiDatabaseModel,
    StateVars,
)
from quik_db.model.validator import model_validator
from quik_db.utils.db_utils import (
    apply_schema_to_query,
    build_sql_proc_params,
    get_default_config,
    has_paging,
    process_params,
)


if TYPE_CHECKING:
    from collections.abc import Callable

    from sqlalchemy.engine import (
        Connection,
        Engine,
        Result,
    )


class Configuration:
    """
    Configuration class for database handlers and attribute instantiation.
    """

    def __init__(
        self,
        config: dict,
        name: str = "",
        handlers: dict[str, Callable | list[Callable]] | None = None,
    ):
        """
        Initializes the database configuration.

        The config is passed to a pydantic model to be validated. This
        is to ensure it has the correct values to create a connection
        or if a feature is enabled that the feature has the values it
        needs to function.

        Args:
            config (dict): The configuration dictionary for the database.
            name (str, optional): The name of the database being called
                in a multidb configiration. Only applies for multidb
                configurations.
            handlers (dict[str, Callable | list[Callable]] | None):
                A dictionary of handlers you can assign to the instance
                class.

        Returns:
            Configuration: An instance of the Configuration
                class.

        Attributes:
            model (BaseDatabaseModel | MultiDatabaseModel): The validated
                pydantic model (BaseModel) with the configuration values.
            execution_handler (Callable): Handler to override the default
                query execution.
            result_handlers (Callable | list[Callable]): Handlers to process
                results from the database.
            error_handlers (Callable | list[Callable]): Handlers to process
                errors from the database or other handlers.
            param_handlers (Callable | list[Callable]): handlers to process
                parameters before executing queries. (e.g. sanitizing)

        """

        self.name = name
        self.result = None
        self.execution_handler = None
        self.result_handlers = []
        self.error_handlers = []
        self.param_handlers = []
        self._exc = ExecutionState.IDLE
        self.vars = StateVars()
        # Validate the configuration against the database models.
        # The returned model is used to set the connection parameters.
        model = model_validator(config)
        _config = model.model_dump()

        # Set the configuration in a multi-database model.
        if model.__class__ == MultiDatabaseModel:
            if name:
                _config = _config["name"]
            elif get_default_config(_config):
                _config = get_default_config(_config)
            elif self.default_config:
                _config = self.default_config
            elif self.default_config_name:
                _config = _config[self.default_config_name]
            else:
                _config = next(iter(_config.items()))

            model = DatabaseModel(**_config)  # pyright: ignore[reportCallIssue]

        self.config = model
        self.qs = self.config.query_settings

        self.conn_build_type = "url" if self.config.url else "params"
        # Set the handlers
        if handlers:
            self.set_handlers(handlers)

    def __repr__(self) -> str:
        """
        Return a string representation of the combined configuration
        and statevars, including only set attributes for config.
        """

        if self.connection:
            self.vars.state = ConnectionState.CONNECTED

        config_data = self.config.model_dump()
        statevars_data = self.vars.model_dump()

        config_data.update(statevars_data)

        return pformat(config_data)

    @property
    def url_string(self) -> str | URL:
        """Return the URL if it is assigned, otherwise return the param URL."""

        return (
            self.config.url
            if self.conn_build_type == "url"
            else self.param_url
        )

    @property
    def param_url(self) -> URL:
        """Return the connection URL built from connection parameters."""
        return URL.create(**self.config.connection_params.model_dump())

    @property
    def url_base(self) -> str:
        """The base string for the connection uri."""
        return "{drivername}://{username}:{password}@{host}:{port}/{database}"

    @property
    def prefix_schema(self):
        """
        Get the schema name from the configuration.

        Returns:
            str: Name of the schema
        """
        schema = self.qs.prefix_schema
        return schema if schema else ""

    @property
    def handler_list(self):
        """Return the list of handlers."""
        return [
            "param_handlers",
            "result_handlers",
            "success_handler",
            "execution_handler",
            "error_handlers",
        ]

    def set_handlers(
        self, handlers: dict[str, Callable | list[Callable]]
    ) -> None:
        """
        Sets the handlers for the database operations.

        This method takes a dictionary of handlers and assigns them to
        the corresponding attributes of the instance. Each handler can
        be a single callable or a list of callables.

        Args:
            handlers (dict[str, Callable | list[Callable]]):
        Returns:
            None
        """

        for handler_name, _handler in handlers.items():
            if handler_name not in self.handler_list:
                raise ValueError(f"Invalid handler name: {handler_name}")
            setattr(self, handler_name, _handler)


class DatabaseConnection(Configuration, ABC):
    """
    Base class for database connections.

    This class provides a standardized configuration framework for
    connectingto various database systems. It supports multi-database
    configurations,allowing you to set a default database if none is
    specified in the configuration file.

    Features:

    - Raw SQL support:
        Helps create raw SQL queries, handle stored procedures, add
        pagination, and provide custom execution and result processing
        handlers.

        >> Example:
        params = {"id": 1}
        query: "SELECT * FROM my_table WHERE id = :id"
        self.execute(query, params)

    - Stored procedure execution:
        Generates a sql to call the procedure based on the database
        dialect. You can provide them with or without the schema or
        database prefix. If you want the prefix to be added
        automatically, you can enable it in query_settings.

        >> Example without prefix (will execute the procedure as is):

        name = "my_procedure"
        self.execute_sp(name, params)

        >> Enable prefixing:
        # YAML
        query_settings:
            schema_prefix: "dbo"

        name = "my_procedure"
        self.execute_sp(name, params)
        # "my_procedure" -> "dbo.my_procedure"

    - Pagination:
        Automatic paging of queries can be set. Set query limit in the
        configuration file for all instances. Paging can also be
        handled at the query level using `limit` and `offset` parameters
        in `execute`. This does not work with stored procedures - only
        raw SQL queries.

        # yaml - affects all instances using this configuration
        dialect: mysql
        url:"mysql+pymysql://user:pass@localhost:3306/mysql"
        query_settings:
            limit: 30

        >> Example (query level):
        self.execute(query, params, offset=offset, limit=limit)

    """

    def __init__(
        self,
        config: dict,
        name: str = "",
        handlers: dict[str, Callable | list[Callable]] | None = None,
    ):
        """
        Initialize the database connection object with the configuration.

        The configuration is validated against the database models to
        ensure that the configuration has the necessary parameters to
        create a connection. The configuration itself supports two
        distinct models: a single database model and a multi-database
        model. Refer to  the documentation on the database models for
        more information.

        Simple example with a single database model:

        - Single Database Model:

            >> with uri
            # yaml
            dialect: mysql
            url:"mysql+pymysql://user:pass@localhost:3306/mysql"
            query_settings:
                limit: 30

            >> with connection params
            dialect: mysql
            query_settings:
                limit: 30
            connection_params:
              host: localhost
              port: 3306
              username: user
              password: pass

        Args:
            handlers (dict[str, Callable | list[Callable]] | None):
                You can assign handlers to the instance class during
                instantiation by passing in a dictionary with the
                handler name as the key and the handler or list of
                handlers as the value.

                e.g.

                myhandlers = {
                    "execution_handler": custom_execution_handler,
                    "result_handlers": [handler1, handler2],
                }
        """
        super().__init__(config, name=name, handlers=handlers)

        self.connection = None
        self._columns = []
        self.exit = None

    def __enter__(self) -> DatabaseConnection:
        """Enter a connection state with the connection object."""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Use the connection's exit method if it exists."""
        if self.connection and hasattr(self.connection, "__exit__"):
            self.connection.__exit__(exc_type, exc_val, exc_tb)
        else:
            self.close()

    def execute_sp(
        self,
        procedure_name: str,
        params: dict[str, Any] | None = None,
        fetch: int | None = None,
    ) -> dict | list:
        """
        Execute a stored procedure with the given parameters.

        Args:
            procedure_name (str): The name of the stored procedure to
                call.
            params (dict[str, Any] | None, optional): A dictionary of
                parameter names and values. Defaults to None.

        Raises:
            ValueError: If the database dialect is not supported.
        """
        self._set_activte_state()
        self.vars.query.append(("execute_sp", procedure_name))
        if not self.connection:
            self.connect()

        dialect = self.config.dialect

        if dialect == "sqlite":
            raise ValueError(errors["sqlite_no_sp"])

        # Apply a schema prefix if enabled
        if not self.vars.schema_applied and self.qs.prefix_procedures:
            procedure_name = self.add_schema_prefix(procedure_name)
            self.vars.schema_applied = True

        if dialect == Dialect.MSSQL and params:
            params_str = ", ".join(f"@{key} = :{key}" for key in params)
        else:
            params_str = build_sql_proc_params(params) if params else ""

        # SQL templates based on dialect
        sql_templates = {
            "postgresql": f"CALL {procedure_name}{params_str}",
            "mysql": f"CALL {procedure_name}{params_str}",
            "mariadb": f"CALL {procedure_name}{params_str}",
            "mssql": f"EXEC {procedure_name} {params_str}",
            "oracle": f"BEGIN {procedure_name}{params_str}; END;",
        }

        _query = sql_templates[dialect]

        return self.execute(query=_query, params=params, fetch=fetch)

    def execute(
        self,
        query: str,
        params: dict[str, str] | tuple[tuple[str, str]] | tuple[str, str] = (),
        fetch: int | None = None,
        offset: int = 0,
        limit: int = 0,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """
        Execute a SQL query with optional parameters and pagination.

        Parameters:
            query (str): SQL query to execute.
            params (dict[str, str] | tuple[str, str]): Query parameters.
            fetch (int): Number of rows to fetch.
            offset (int): Row offset for pagination.
            limit (int): Limit of rows to fetch.

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Query result as a
            list of dictionaries or a dictionary.
        """
        self._set_activte_state()
        result = None
        # Check for connection
        if not self.connection:
            self.connect()

        sv = self.vars
        # Add query for tracking
        sv.query.append(("execute_input", query))
        # Track the input
        sv.params.append(("input", params))
        # Prepare query with schema and pagination
        if not self.vars.schema_applied and self.qs.prefix_queries:
            query = self.add_schema_prefix(query, "query")

        if offset or limit:
            query = self.add_paging(query, offset, limit)
            sv.query.append(("paging", query))
        # Process parameters and update query params state
        if self.param_handlers and params:
            params = process_params(params, self.param_handlers)
            sv.params.append(("processed", params))
        try:
            if callable(self.execution_handler):
                self.result = self.execution_handler(query, params, fetch)  # pylint: disable=not-callable
                # Execute the query
            else:
                result = self.result = self.connection.execute(
                    text(query), params
                )

                # Check for results
                if not result.returns_rows:
                    return result

                sv.columns = result.keys()

                # Get the fetch value based on fetch settings
                fetch_size = self._get_fetch_value(fetch)

                if fetch_size is not None:
                    result = self.fetch_results(fetch_size)

            return self.process_results(result)

        except* Exception as e:  # pylint: disable=broad-except
            self.vars.state = ConnectionState.ERROR
            if self.error_handlers:
                exception = handler(self.error_handlers, e)
                if isinstance(exception, Exception):
                    raise exception from e
            raise e

        finally:
            self._exc = ExecutionState.IDLE
            if self.vars.state != ConnectionState.ERROR:
                self.commit()

    def add_paging(
        self,
        query: str,
        offset: int = 0,
        limit: int = 0,
    ) -> str:
        """
        Adds paging (limit and offset) to an SQL query based on the
        dialect.

        Args:
            query (str): The SQL query to modify.
            offset (int): Number of records to skip.
            limit (int): Maximum number of records to return.

        Returns:
            str: The SQL query with paging added, or the original if
            paging is present.
        """

        limit = self.get_limit(limit)

        if has_paging(query):
            return query

        # Retrieve the statevariables
        var = self.vars

        offset = var.offset or offset

        # Limit can be set in the config as well
        limit = var.limit or limit or self.config.params.get("limit", 0)

        paging = ""

        if self.config.dialect in {"mysql", "mariadb", "postgres", "sqlite"}:
            # If only offset is provided, allow 'LIMIT ALL'
            if limit > 0:
                paging = f"LIMIT {limit} OFFSET {offset}"
            elif offset > 0:
                paging = f"LIMIT ALL OFFSET {offset}"

        elif self.config.dialect in {"sql server", "oracle"}:
            # SQL Server and Oracle require 'FETCH NEXT' with OFFSET
            if limit > 0:
                paging = f"OFFSET {offset} ROWS FETCH NEXT {limit} ROWS ONLY"
            elif offset > 0:
                paging = f"OFFSET {offset} ROWS FETCH NEXT ALL ROWS ONLY"

        return f"{query} {paging}" if paging else query

    def process_results(
        self, result: dict | Result, query_type: QueryType = QueryType.QUERY
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """
        Process the results from a database query or return a success
        message for DML (Data Manipulation Language) queries.

        Args:
            result: The result object or dictionary from the database query.
            query_type: The type of query being processed; defaults to `QUERY`.

        Returns:
            A list of dictionaries representing query results or a success
            message with the affected row count for DML queries.
        """
        d = self.config.dialect

        if query_type == QueryType.PROCEDURE and d in (
            Dialect.ORACLE,
            Dialect.POSTGRESQL,
        ):
            rows: int = result.rowcount
            return {
                "message": f"Success. Rows affected {rows}",
                "rows_affected": rows,
            }

        if self.result_handlers:
            return handler(self.result_handlers, result)
        return result

    def fetch_results(self, fetch_size: int) -> dict[str, Any] | list:
        """Fetch results from the result object as mappings."""

        result = self.result
        if not result:
            raise ValueError(errors["no_result"])

        if self.config.result_as_mappings:
            if fetch_size == 0:
                return result.mappings().all()
            return result.mappings().fetchmany(fetch_size)

        if fetch_size == 0:
            return result.all()

        return result.fetchmany(fetch_size)

    def add_schema_prefix(
        self, query: str, query_type: Literal["procedure", "query"] = "query"
    ) -> str:
        """Add the schema to stored procedure/function names."""

        if query_type == "procedure":
            if "." in query:
                warn(errors["prefix_error"], stacklevel=2)
            return f"{self.prefix_schema}.{query}"

        return apply_schema_to_query(
            query, self.prefix_schema, self.config.dialect
        )

    def limit(self, limit: int):
        self._set_activte_state()
        "Assigns a limit to the query."
        self.vars.limit = limit
        return self

    def offset(self, offset: int):
        self._set_activte_state()
        """Assign an offset to the query."""
        self.vars.offset = offset
        return self

    def fetch(self, fetch_size: int):
        self._set_activte_state()
        """Fetch results from the result object."""
        self.vars.fetch_size = fetch_size

        return self

    def rollback(self) -> None:
        """Rollback the transaction."""
        if self.connection:
            self.connection.rollback()
        else:
            raise ConnectionError(errors["no_connection"])

    def commit(self) -> None:
        """Commit the transaction."""
        if self.connection:
            self.connection.commit()
        else:
            raise ConnectionError(errors["no_connection"])

    def close(self) -> None:
        """
        Close the connection to the database.

        Subclass and override if conn.close() is not supported by the
        database.
        """
        self.vars.state = ConnectionState.DISCONNECTED

        if self.connection and self.vars.state not in [
            ConnectionState.DISCONNECTED,
            ConnectionState.ERROR,
        ]:
            self.connection.close()
            self.connection = None

    def _set_activte_state(self) -> None:
        """Check and set the execution state"""

        if self._exc == ExecutionState.IDLE:
            self._exc = ExecutionState.ACTIVE
            # Reset the statevars
            self.vars = StateVars()

    def _get_fetch_value(self, fetch_size: int | None):
        """
        Return the fetch value for the query.
        """
        fetch_size = fetch_size or self.vars.fetch_size

        if fetch_size is None and self.qs.enable_fetch:
            fetch_size = 0
        return fetch_size

    @abstractmethod
    def connect(self) -> Connection | Session | scoped_session:
        """
        Open, set, and return the connection.
        """
        raise NotImplementedError(errors["no_connect_method"])


class SqlAlchemyConnection(DatabaseConnection):
    """Create a SqlAlchemyDatabase object from a configuration file."""

    def __init__(
        self,
        config: dict,
        name: str = "",
        connection_type: Literal["scoped", "session", "direct"] = "direct",
        session: Session | scoped_session | None = None,
    ) -> None:
        """
        Initialize the database connection.

        Args:
            config (dict): Configuration dictionary for the database.
            name (str): Name of the database. Defaults to empty string.
            connect (bool): If True, connect on instantiation. Defaults
                to False.
            connection_type (Literal['scoped', 'session', 'direct']):
                Type of connection ('scoped', 'session', 'direct').
                Defaults to 'direct'.
            session (sessionmaker | scoped_session | None): Existing
                session factory.
        """

        if not session and not config:
            raise ValueError(errors["no_config"])

        super().__init__(config, name=name)

        self._session_factory = session if session else None
        self.connection_type = connection_type

        if session:
            self.connect()

    @property
    def engine(self) -> Engine:
        """Create a SQLAlchemy engine."""
        engine_options = self.config.options if self.config.options else {}
        return create_engine(self.url_string, **engine_options)

    def session(self) -> SqlAlchemyConnection:
        """
        Return a new session from the sessionmaker while maintaining
        the SqlAlchemyConnection instance configuration
        """

        if not self._session_factory:
            raise ValueError(errors["no_session"])

        return SqlAlchemyConnection(
            config=self.config,
            name=self.name,
            connection_type=self.connection_type,
            session=self._session_factory(),  # pyright: ignore[reportCallIssue]
        )

    @override
    def connect(self) -> Connection | Session | scoped_session:
        """Open, set, and return the connection."""

        if self.connection:
            return self.connection

        # create a connection based off existing session
        if self._session_factory:
            self.connection = self._session_factory()

        conn_type = self.connection_type

        if conn_type == "direct":
            self.connection = self.engine.connect()
        else:
            # Create a session and store it
            self._session_factory = self._create_session_factory()
            # Create a connection off the session
            self.connection = self._session_factory()

        if not self.connection:
            raise ConnectionError(errors["no_connection"])

        self.vars.state = ConnectionState.CONNECTED
        return self.connection

    def _create_session_factory(
        self,
    ) -> sessionmaker | scoped_session:
        """Helper method to create a session or connection."""
        if self.connection_type == "session":
            return sessionmaker(bind=self.engine)
        if self.connection_type == "scoped":
            return scoped_session(sessionmaker(bind=self.engine))
        raise ValueError(errors["invalid_session_type"])

    @override
    def close(self) -> None:
        """Close the connection to the database."""
        if self.connection and self.vars.state not in [
            ConnectionState.DISCONNECTED,
            ConnectionState.ERROR,
        ]:
            if isinstance(self._session_factory, scoped_session):
                self.connection.remove()
            else:
                self.connection.close()

            self.connection = None
            self.vars.state = ConnectionState.DISCONNECTED



"Import dictionaries"

from quik_db.dictionaries.db_errors import db_errors
from quik_db.dictionaries.model_errors import (
    db_model_errors,
    db_validation_errors,
)
from quik_db.dictionaries.model_fields import db_model_fields
from quik_db.dictionaries.resources import drivers, url_keys


__all__ = [
    "db_errors",
    "db_model_errors",
    "db_model_fields",
    "db_validation_errors",
    "drivers",
    "url_keys",
]



"""
Collection of dictionaries used to store messages related to database models,
configuration, and errors.
"""

from quik_db.utils.formatters import flatten_dict


connection_params_fields = {
    "drivername": {
        "title": "Database Driver",
        "description": (
            "Specifies the driver for the database connection, which may vary "
            "across database dialects or when ODBC support is enabled."
        ),
        "example": "mysql+pymysql",
    },
    "host": {
        "title": "Database Host",
        "description": "The hostname or IP address of the database server.",
        "example": "127.0.0.1",
    },
    "port": {
        "title": "Database Port",
        "description": "The port number that the database server is listening on.",
        "example": 5432,
    },
    "username": {
        "title": "Database Username",
        "description": "The username for authenticating the database connection.",
        "example": "user123",
    },
    "password": {
        "title": "Database Password",
        "description": "The password associated with the database username.",
        "example": "pass123",
    },
    "database": {
        "title": "Database Name",
        "description": "The name of the specific database to connect to.",
        "example": "mydatabase",
    },
    "query": {
        "title": "URL Query Parameters",
        "description": (
            "Optional query parameters to include in the connection URL, "
            "formatted as key-value pairs (e.g., ?param1=value1&param2=value2)."
        ),
        "example": {"param1": "value1", "param2": "value2"},
    },
}

connection_params_fields = flatten_dict(connection_params_fields)

query_settings_fields = {
    "enable_fetch": {
        "title": "Enable Fetch",
        "description": (
            "Enabled by default. Fetches results from the Result object. "
            "Set to False to disable fetching and return the Result object."
        ),
    },
    "prefix_schema": {
        "title": "Schema Prefix",
        "description": (
            "The prefix format to add to query names, stored procedures, "
            "or functions if they do not already have a schema prefix."
        ),
    },
    "prefix_queries": {
        "title": "Prefix Queries",
        "description": "If set to True, applies the schema prefix to queries.",
    },
    "prefix_procedures": {
        "title": "Prefix Procedures",
        "description": (
            "If set to True, applies the schema prefix to stored procedures "
            "or function names."
        ),
    },
    "limit": {
        "title": "Query Limit",
        "description": (
            "Limits the number of rows returned in raw queries, implementing "
            "paging. Setting this to 0 disables paging."
        ),
        "example": 100,
    },
}

query_settings_fields = flatten_dict(query_settings_fields)

db_model_fields = {
    "connection_params": {
        "title": "Connection Parameters",
        "description": (
            "Parameters for connecting to a database. Used in replacement"
            " of the `url` field"
        ),
    },
    "default_config": {
        "title": "Default Config",
        "description": (
            "Specifies the default configuration to use when a database name "
            "is not provided during DatabaseConnection initialization."
        ),
        "example": "default_db_config",
    },
    "default_config_name": {
        "title": "Default Config Name",
        "description": (
            "The key name of the default database configuration, used to "
            "reference this configuration when multiple databases are "
            "configured in a MultiDatabase model"
        ),
        "example": "mydatabase",
    },
    "dialect": {
        "title": "Database Dialect",
        "description": (
            "The type of database being connected to. Supported values include "
            "mysql, mariadb, mssql, postgresql, oracle, and sqlite."
        ),
        "example": "postgresql",
    },
    "description": {
        "title": "Description",
        "description": "A description field for self documentation.",
    },
    "default": {
        "title": "Is Default",
        "description": "If set to True, marks this database configuration as "
        "the default database in a multi-database configuration.",
    },
    "options": {
        "title": "Connection Options",
        "description": (
            "Additional connection options, such as echo settings, pool size, "
            "or connection arguments."
        ),
        "example": {"echo": True, "pool_size": 5},
    },
    "query_settings": {
        "title": "Query Settings",
        "description": "Settings for query execution and result handling.",
    },
    "result_as_mappings": {
        "title": "Return as Mappings",
        "description": "If set to True, returns query results as a list of dictionaries.",
    },
    "schema_name": {
        "title": "Schema Name",
        "description": "The schema_name used. May not be applicable for some dialects",
        "example": "public",
    },
    "url": {
        "title": "Database URL",
        "description": (
            "A connection string or DSN name. For ODBC, this refers to the DSN "
            "name."
        ),
        "example": "postgresql://user:password@localhost/dbname",
    },
}
db_model_fields = flatten_dict(db_model_fields)



from __future__ import annotations

import re
import string


def clean_data(
    data: str | dict | list, lowercase: bool = False
) -> str | dict | list:
    """
    Cleans data based on type. Removes whitespaces and can convert to lowercase
    if specified.

    :param data: data to be cleaned and lowered if True
    :type data: dict, list, str
    :param lowercase: Lowercases nested structure, defaults to False
    :type lowercase: bool, optional
    :return: data
    :rtype: dict, list, str
    """
    if isinstance(data, dict):
        return {
            clean_data(k, lowercase): clean_data(v, lowercase)
            for k, v in data.items()
        }
    elif isinstance(data, list):
        return [clean_data(item, lowercase) for item in data]
    elif isinstance(data, str):
        return re.sub(r"\s+", "", data.lower() if lowercase else data)
    else:
        return data


def capitalize(str: str) -> str:
    "Capitalizes the first character within a group"
    return re.sub(
        r"(^|[/!?]\s+)([a-z])", lambda m: m.group(1) + m.group(2).upper(), str
    )


def lowercase_nested_data(data: str | dict | list) -> str | dict | list:
    """
    Lowercases all keys in nested dictionaries and lists.

    :param data: data to be lowercased
    :type data: str, dict, list
    :return: data
    :rtype: str, dict, list
    """
    if isinstance(data, dict):
        return {k.lower(): lowercase_nested_data(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [lowercase_nested_data(item) for item in data]
    elif isinstance(data, str):
        return data.lower()
    else:
        return data


class SafeDict(dict):
    def __missing__(self, key):
        return "{" + key + "}"


def safe_format(format_string: str, **kwargs) -> str:
    """Format the string with available keys in the dictionary."""
    formatter = string.Formatter()
    safe_dict = SafeDict(**kwargs)
    return formatter.vformat(format_string, (), safe_dict)


def clean_dict(dictionary: dict) -> dict:
    """Removes None values from the dictionary."""

    # if a list, clean each item in the list
    if isinstance(dictionary, list):
        return [clean_dict(item) for item in dictionary]

    # if not a dictionary or a tuple, just return it
    if not isinstance(dictionary, dict):
        return dictionary

    return {
        key: clean_dict(val)
        for key, val in dictionary.items()
        if val is not None
    }


def flatten_str(message: str) -> str:
    """Removes line breaks and excess spaces from the string"""
    return " ".join(message.split())


def flatten_dict(message: dict) -> dict:
    """Flattens strings in a dictionary"""
    for key, value in message.items():
        if isinstance(value, dict):
            message[key] = flatten_dict(value)
        elif isinstance(value, list):
            message[key] = [
                (
                    flatten_dict(item)
                    if isinstance(item, dict)
                    else flatten_str(item)
                )
                for item in value
            ]
        elif isinstance(value, str):
            message[key] = flatten_str(value)
    return message


def fix_format_args(query: str) -> None:
    """
    Checks if the SQL query string contains incorrectly formatted
    arguments, such as {arg}, %s, %d, f-strings, or string
    concatenation.

    Args:
        query (str): The SQL query string.

    Returns:
        str: Fixed query string compatible with the text() construct
    """
    # Define patterns for incorrect formatting
    incorrect_patterns = [
        r"\{.*?\}",  # Curly braces for format() or f-strings
        r"%s",
        r"%d",
        r"%\w",  # Old-style % formatting
        r'f".*?"',  # f-strings (Python 3.6+)
        r"f\'.*?\'",  # f-strings with single quotes
    ]

    # Check for each incorrect pattern
    for pattern in incorrect_patterns:
        if re.search(pattern, query):
            return True  # Return True if any incorrect format is found

    # If none of the incorrect formats are found, return False
    return False



"""
Validation Module for database models.

This module provides a validator function for validating any db_models
against a given database configuration. It uses Pydantic for validation and

"""

from __future__ import annotations

from typing import Any

from pydantic import ValidationError

from quik_db.model.exceptions import ModelValidationError
from quik_db.model.models import DatabaseModel, MultiDatabaseModel


db_models = {
    "databasemodel": DatabaseModel,
    "multidatabasemodel": MultiDatabaseModel,
}


def model_validator(
    db_config: dict[str, Any],
    models: dict[str, Any] | None = None,
    model_name: str = "",
) -> DatabaseModel | MultiDatabaseModel:
    """
    Validates the given database configuration against the provided models.
    Args:
        db_config (dict[str, Any]): The database configuration to validate.
        models (dict[str, type[BaseModel]]): A dictionary of model names to
            model classes.
        model_name (str | None, optional): The name of a specific model to
            validate against. If None, all models will be validated.
    Returns:
        BaseModel: The validated model instance.
    Raises:
        ModelValidationError: If validation fails for all models or the
        specified model.
    """

    model_errors = {}
    models = models if models else db_models
    model_name = model_name.lower() if model_name else ""

    # Validation function. Tracks errors for each model.
    def validate_model(
        model_class: type[DatabaseModel | MultiDatabaseModel],
    ) -> MultiDatabaseModel | DatabaseModel | None:
        try:
            return model_class(**db_config)
        except ValidationError as e:
            model_errors[model_class.__name__] = e.errors()
            return None

    # Validate the specified model or all models
    if model_name:
        validated_model = validate_model(models[model_name])
        if validated_model:
            return validated_model
    else:
        for model_class in models.values():
            validated_model = validate_model(model_class)
            if validated_model:
                return validated_model

    raise ModelValidationError(model_errors)



"Other resource dictionaries"

drivers = {
    "postgresql": "postgresql+psycopg2",
    "mysql": "mysql+pymysql",
    "sqlite": "sqlite",
    "oracle": "oracle+oracledb",
    "mssql": "mssql+pymssql",
    "mariadb": "mariadb+pymysql",
}

url_keys = (
    "drivername",
    "username",
    "password",
    "host",
    "port",
    "database",
    "query",
)



# SPDX-FileCopyrightText: 2024-present Victor Nguyen <victor.win86@gmail.com>
#
# SPDX-License-Identifier: MIT
# ruff: noqa: F401
# flake8: noqa: E501
# pyright: ignore[reportUnusedImport, W0404]
# pylint: disable=W0404
from __future__ import annotations

from quik_db.core.database import (
    DatabaseConnection,
    SqlAlchemyConnection,
    SqlAlchemyConnection as Connection,
)
from quik_db.model import (
    DatabaseModel,
    ModelValidationError,
    MultiDatabaseModel,
    model_validator,
)



"""
Database Validation Models used for validating configurations
"""

# flake8: noqa: E501
# pyright: reportArgumentType=false
# pylint: disable=C0115
# ruff: noqa: E712
from __future__ import annotations

from typing import Any, ClassVar

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    ValidationError,
    model_validator,
)
from pydantic_core import InitErrorDetails

from quik_db.dictionaries import drivers
from quik_db.dictionaries.model_fields import (
    connection_params_fields,
    db_model_fields,
    query_settings_fields,
)
from quik_db.model.enums import ConnectionState, Dialect
from quik_db.model.exceptions import create_error_msg
from quik_db.model.metadata import update_model_metadata
from quik_db.utils.db_utils import get_default_db_count


class ModelConfig(BaseModel):
    """
    Set the default configuration for the BaseModel
    """

    model_config = ConfigDict(
        extra="forbid",
        validate_assignment=True,
    )


class StateVars(BaseModel):
    """
    State variables for the database connection. This model is not
    designed  to be instantiated directly but rather serves as a
    container for state information related to the database execution.
    """

    model_config = ConfigDict(extra="allow", validate_assignment=True)

    state: ConnectionState = ConnectionState.DISCONNECTED
    connection_type: str = ""
    schema_applied: bool = False
    fetch_size: int | None = None
    columns: str | list = []
    page: int = 0
    offset: int = 0
    row_count: int = 0
    query: list = Field(default_factory=list)
    params: list = Field(default_factory=list)
    results: list = Field(default_factory=list)

    def update(
        self, key: str | None = None, value: Any = None, **kwargs: Any
    ) -> None:
        """
        Update state variables with validation.

        Args:
            key (str, optional): A single attribute name to set.
            value (Any, optional): The value for the single attribute.
            kwargs (Any): Multiple attribute-value pairs to set at once.
        """
        if key and value is not None:
            # Single key-value update
            setattr(self, key, value)
        else:
            # Multiple updates via kwargs
            for k, v in kwargs.items():
                if hasattr(self, k) and isinstance(getattr(self, k), list):
                    getattr(self, k).append(v)
                else:
                    setattr(self, k, v)


@update_model_metadata(connection_params_fields)
class ConnectionParams(ModelConfig):
    """
    Connection parameters for the database.

    This model is used to store the necessary information to connect to
    a database.The driver is optional only if odbc is not utilized. If
    url is present in the model, it will override connection parameters.
    Review the model_validator in DatabaseModel for more information.

    """

    model_config = ConfigDict(extra="forbid")

    drivername: str = ""
    host: str = ""
    port: str | int = ""
    username: str = ""
    password: str = ""
    database: str = ""
    query: dict = Field(default_factory=dict)


@update_model_metadata(query_settings_fields)
class QuerySettings(ModelConfig):
    """
    Query model for database operations.

    """

    enable_fetch: bool = True
    prefix_schema: str = ""
    prefix_queries: bool = False
    prefix_procedures: bool = False
    limit: int = 0

    def __init__(self, **data):
        """Sets the default values when enabling prefix_schema."""

        super().__init__(**data)

        p_queries = data.get("prefix_queries", None)
        p_procedures = data.get("prefix_procedures", None)

        # Set the defaults to be True if not defined
        if self.prefix_schema and (p_queries is None and p_procedures is None):
            self.prefix_queries = True
            self.prefix_procedures = True


@update_model_metadata(db_model_fields)
class DatabaseModel(BaseModel):
    """
    Single database configuration model.

    The model forces a configuration to align with the expected schema
    to ensure that the database connection has the necessary parameters
    to create a connection.

    """

    model_config = ConfigDict(extra="ignore", use_enum_values=True)
    # Do not use. Tracks execution state.
    # Class attributes
    default_config: ClassVar[str] = ""
    default_config_name: ClassVar[str] = ""
    reserved_attrs: ClassVar[list] = Field(default_factory=list)

    dialect: Dialect
    description: str = ""
    default: bool = False
    url: str = ""
    connection_params: ConnectionParams = Field(
        default_factory=ConnectionParams
    )
    schema_name: str = ""
    options: dict[str, Any] = Field(default_factory=dict)
    result_as_mappings: bool = False
    query_settings: QuerySettings = Field(default_factory=QuerySettings)

    @model_validator(mode="before")
    @classmethod
    def field_validation(cls, values: dict) -> dict:
        """
        Validates the fields of a database configuration dictionary.
        Args:
            cls: The class that calls this method.
            values (dict): The dictionary containing database configuration.
        Returns:
            dict: The validated and possibly modified configuration dictionary.
        Raises:
            TypeError: If the input `values` is not a dictionary.
            ValidationError: If any validation checks fail.
        Validation Checks:
            - Ensures `values` is a dictionary.
            - Checks for the presence and validity of the `dialect` key.
            - Validates connection parameters (`username`, `password`,
            `host`, `driver`).
            - Validates query settings related to schema prefix.
            - For SQLite, ensures the presence of the `path` key.
            - For other dialects, ensures either `url` or connection
            parameters are provided.
            - Sets a default driver if not provided.
        Note:
            - Raises `ValidationError` with detailed error messages if
                any validation fails.
        """

        errors = []
        if not isinstance(values, dict):
            errors.append(create_error_msg("invalid_value_type"))
        else:
            dialect = values.get("dialect")
            url = values.get("url")
            # Check for missing dialect
            if not dialect:
                errors.append(create_error_msg("missing_dialect"))
            if dialect not in Dialect:
                errors.append(create_error_msg("invalid_dialect"))

            # Connection Parameters
            params = values.get("connection_params") or {}
            username = params.get("username")
            password = params.get("password", "")
            host = params.get("host")
            drivername = params.get("driver") if params else None

            # Query Settings
            qs = values.get("query_settings", {})
            schema_prefix = qs.get("prefix_schema")
            prefix_queries = qs.get("prefix_queries", None)
            prefix_procedures = qs.get("prefix_procedures", None)

            if schema_prefix:
                if prefix_queries is False and prefix_procedures is False:
                    errors.append(create_error_msg("invalid_prefix_settings"))

            if not schema_prefix and (prefix_queries or prefix_procedures):
                errors.append(create_error_msg("invalid_prefix_settings"))

            # Sqlite only requires the url (path to .db file)
            if dialect == "sqlite":
                if not url or ".db" not in url:
                    errors.append(create_error_msg("missing_sqlite_url"))
                if params:
                    errors.append(create_error_msg("invalid_sqlite_params"))
            else:
                if url and params:
                    errors.append(create_error_msg("invalid_url_and_params"))
                if not (url or params):
                    errors.append(create_error_msg("missing_url_or_params"))
                if params:
                    if not host:
                        errors.append(create_error_msg("invalid_params"))
                    if username and not password:
                        errors.append(create_error_msg("missing_password"))
                if not url and params and not drivername:
                    values["connection_params"]["drivername"] = drivers[
                        dialect
                    ]

        if errors:
            _errors = [InitErrorDetails(err) for err in errors]
            raise ValidationError.from_exception_data(
                title=cls.__name__,
                line_errors=_errors,
            )

        # Set the default driver

        return values


class MultiDatabaseModel(RootModel[str | dict[str, DatabaseModel]]):
    """
    Model for Multidatabase configurations.

    Conifigurations that are defined
    with a root key (database name) and only have 1 value (DatabaseModel)
    will pass validation.

    e.g.

    mydatabase1:
      dialect: mysql
      url:mysql://user:pass
    mydatabase2:
        dialect: sqlite
        url:sqlite:///path/to/db.sqlite
    """

    @model_validator(mode="before")
    @classmethod
    def validate_schema(cls, values: dict) -> dict:
        """
        Validate the schema of the provided dictionary.

        This method checks if the input is a dictionary and validates
        the structure of its nested dictionaries. It also ensures that
        there is at most one default database specified.

        Args:
            values (dict): The dictionary to validate.

        Returns:
            dict: The validated dictionary.

        Raises:
            TypeError: If the root element is not a dictionary.
            ValidationError: If there are any validation errors, such as
                invalid nesting or multiple default databases.

        """
        errors = []

        for v in values.values():
            if v is None or not isinstance(v, dict):
                error = create_error_msg("invalid_nesting")
                if error not in errors:
                    errors.append(error)
        # Checking for multiple defaults=
        if get_default_db_count(values) > 1:
            errors.append(create_error_msg("multiple_default_db"))

        if errors:
            _errors = [InitErrorDetails(err) for err in errors]
            raise ValidationError.from_exception_data(
                title=cls.__name__,
                line_errors=_errors,
            )

        # Return with meta
        return values



"Metadata updator for Pydantic models."
# pylint: disable=C0301
# flake8: noqa E501
# ruff: noqa: E501

from pydantic import Field


def update_model_metadata(metadata_dict: dict):
    """
    Updates the fields of a Pydantic model class with metadata from
    a dictionary.

    Args:
        cls (BaseModel): A Pydantic model class.
        metadata_dict (dict): A dictionary containing field metadata,
            where each key is a field name and each value is a
            dictionary of metadata.
    """

    def decorator(cls):
        for field_name, metadata in metadata_dict.items():
            if field_name in cls.model_fields:
                cls.model_fields[field_name] = Field(
                    title=metadata.get("title"),
                    description=metadata.get("description"),
                    json_schema_extra={"example": metadata.get("example")},
                )
        cls.model_rebuild()
        return cls

    return decorator



"Model Validation Errors"

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

from quik_db.dictionaries import db_validation_errors


if TYPE_CHECKING:
    from pydantic_core import ErrorDetails


class ModelValidationError(Exception):
    """
    Exception for model validation errors.

    This exception is raised when all models fail to validate the
    configuration.
    """

    def __init__(self, model_errors: dict[str, list[dict]]) -> None:
        """
        Initialize with a dictionary of errors.

        :param model_errors: Dictionary of error messages values from
        ValidationError.errors().

        """
        self.errors = model_errors
        self.models = list(model_errors.keys())
        super().__init__(self._generate_message())

    def _generate_message(self) -> str:
        """
        Generate a comprehensive error message combining all
        model errors.
        """
        messages = []
        model_count = len(self.errors)
        plural = "s" if model_count != 1 else ""
        messages.append(f"{model_count} model{plural} tested for validation.")

        for model, errors in self.errors.items():
            error_count = len(errors)
            messages.append(
                f"\nModel '{model}' encountered {error_count} error(s):",
            )
            for error in errors:
                loc = ".".join(map(str, error.get("loc", [])))
                msg = get_error_msg(error)
                type_ = error.get("type", "")
                # Initialize error message
                error_msg = (
                    f"  - Location: {loc}\n    "
                    f"Message: {msg}\n    "
                    f"Type: {type_}"
                )
                # Add context details
                for k, v in error.get("ctx", {}).items():
                    error_msg += f"\n    {k}: {v}"

                messages.append(error_msg)

        # Summary information
        total_errors = sum(len(errs) for errs in self.errors.values())
        messages.append(f"\nTotal Errors: {total_errors}")
        return "\n".join(messages)


def get_error_msg(error: dict | ErrorDetails) -> str:
    """
    Extract and return an error message from a given error dictionary.

    Args:
        error (dict): A dictionary containing error information.

    Returns:
        str: The extracted error message. If the "ctx" dictionary
            contains an "error" key, its value is returned. Otherwise,
            the value of the "msg" key in the error dictionary is
            returned. If neither is present, an empty string is returned.

    """
    ctx = error.get("ctx", {})
    if ctx and ctx.get("error"):
        return ctx.pop("error")
    return error.get("msg", "")


class SettingValidationError(Exception):
    """Exception to hold a list of validation errors."""

    def __init__(self, errors: list):
        super().__init__("Validation errors occurred.")
        self.errors = errors  # Store the list of errors as-is


def create_error_msg(key: str, ctx: dict | str | None = None) -> dict:
    """
    Retrieve and update an error dictionary for the given key.

    Retrieves the error dictionary from `db_validation_errors` and, if
    `ctx` is provided, updates the "ctx" section in the error dictionary
    with values from `ctx`.

    Args:
        key (str): The key to retrieve the error dictionary.
        ctx (dict, optional): Dictionary with keys to update the "ctx"
            field within the error dictionary.

    Returns:
        dict: Updated error dictionary with None values removed.
    """
    error = deepcopy(db_validation_errors[key])

    # Update the context if provided
    if ctx:
        if isinstance(ctx, str):
            ctx = {"value": ctx}
        if "ctx" not in error:
            error["ctx"] = {}
        error["ctx"].update(ctx)

    # Return only set values
    return {k: v for k, v in error.items() if v is not None}



"""
This module is for handlers you can use in the DatabaseConnection objects.
"""

# ruff: noqa: E501
# flake8: noqa: E501
from __future__ import annotations

import re
from typing import Any, Callable

from dateutil import parser
from sqlalchemy.engine.row import Row


def handler(
    handlers: Callable | list[Callable],
    data: Any,
) -> Any:
    """
    Process the data with the given handler(s).

    Args:
        data: The data to process.
        handler: A single handler or a list of handlers.
        out: Provide the default output for the handler(s) if there is no
            expected return value.

    Returns:
        The processed data.
    """
    if isinstance(handlers, list):
        for h in handlers:
            data = h(data)
    else:
        data = handlers(data)
    return data


class DateFormatter:
    """
    Class to recursively format date strings in a data structure.

    Attributes:
        date_format (str): The format to convert date strings into.
        date_patterns (list[str]): list of regex patterns to identify
            date strings.
        compiled_patterns (list[Pattern]): Compiled regex patterns.
    """

    def __init__(self, date_format: str = "%Y-%m-%d") -> None:
        """
        Initializes the DateFormatter with a specific date format.

        Args:
            date_format (str): The desired date format for output
                strings.
        """
        self.date_format = date_format
        # Define regex patterns for date/datetime strings
        self.date_patterns = [
            r"\b\d{4}-\d{1,2}-\d{1,2}\b",  # YYYY-M-D or YYYY-MM-DD
            r"\b\d{1,2}/\d{1,2}/\d{2,4}\b",  # M/D/YYYY or MM/DD/YY
            r"\b\d{1,2}-\d{1,2}-\d{2,4}\b",  # D-M-YYYY or DD-MM-YY
            r"\b\d{4}/\d{1,2}/\d{1,2}\b",  # YYYY/M/D or YYYY/MM/DD
            r"\b\d{4}-\d{1,2}-\d{1,2} \d{2}:\d{2}:\d{2}\b",  # YYYY-M-D HH:MM:SS
            r"\b\d{4}/\d{1,2}/\d{1,2} \d{2}:\d{2}:\d{2}\b",  # YYYY/M/D HH:MM:SS
            r"\b\d{1,2} \w{3} \d{4}\b",  # D Mon YYYY
            r"\b\w{3} \d{1,2}, \d{4}\b",  # Mon D, YYYY
        ]

        self.compiled_patterns = [re.compile(p) for p in self.date_patterns]

    def is_date_string(self, s: str) -> bool:
        """
        Checks if a string matches any of the date regex patterns.

        Args:
            s (str): The string to check.

        Returns:
            bool: True if the string is a date, False otherwise.
        """
        return any(pattern.fullmatch(s) for pattern in self.compiled_patterns)

    def process_data(self, data: Any) -> Any:
        """
        Recursively processes the data to format date strings.

        Args:
            data (Any): The data to process (dict, list, str, or other).

        Returns:
            Any: The processed data with formatted date strings.
        """
        if isinstance(data, dict):
            return {k: self.process_data(v) for k, v in data.items()}
        elif isinstance(data, list):
            return [self.process_data(item) for item in data]
        elif isinstance(data, str):
            if self.is_date_string(data):
                try:
                    dt = parser.parse(data)
                    return dt.strftime(self.date_format)
                except (ValueError, OverflowError):
                    return data
            else:
                return data
        else:
            return data

    def __call__(self, data: Any) -> Any:
        """
        Allow the instance to be called like a function.

        Args:
            data (Any): The data to process.

        Returns:
            Any: The processed data.

        """
        return self.process_data(data)


def single_row_to_dict(data: list[dict]) -> dict:
    """Manage the conversion of a single row of data to a dictionary."""
    if len(data) == 1 and isinstance(data, list):
        return data[0]
    return data


def result_to_dict(
    result: Row | list[Row],
) -> dict[str, Any] | list[dict[str, Any]]:
    """
    Converts SQLAlchemy result to a dictionary or list of dictionaries.

    Args:
        result: A single SQLAlchemy Row or a list of Rows.

    Returns:
        A dictionary or a list of dictionaries.
    """
    if isinstance(result, Row):
        return dict(result)
    return [dict(row) for row in result]



# db_connection.pyi

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

from sqlalchemy import URL, Engine
from sqlalchemy.engine import Connection, Result
from sqlalchemy.orm import (
    Session,
    scoped_session,
    sessionmaker,
)

from quik_db.model.enums import ExecutionState, QueryType
from quik_db.model.models import (
    DatabaseModel,
    MultiDatabaseModel,
    QuerySettings,
    StateVars,
)


if TYPE_CHECKING:
    from collections.abc import Callable


class Configuration:
    name: str
    result: Optional[Any]
    execution_handler: Optional[Callable[..., Any]]
    result_handlers: List[Callable[..., Any]]
    error_handlers: List[Callable[..., Any]]
    param_handlers: List[Callable[..., Any]]
    _exc: ExecutionState
    vars: StateVars
    config: Union[DatabaseModel, MultiDatabaseModel]
    qs: QuerySettings
    conn_build_type: Literal["url", "params"]

    def __init__(
        self,
        config: Dict[str, Any],
        name: str = "",
        handlers: Optional[
            Dict[str, Union[Callable[..., Any], List[Callable[..., Any]]]]
        ] = None,
    ) -> None: ...

    def __repr__(self) -> str: ...

    @property
    def url_string(self) -> Union[str, URL]: ...

    @property
    def param_url(self) -> URL: ...

    @property
    def url_base(self) -> str: ...

    @property
    def prefix_schema(self) -> str: ...

    @property
    def handler_list(self) -> List[str]: ...

    def set_handlers(
        self,
        handlers: Dict[
            str, Union[Callable[..., Any], List[Callable[..., Any]]]
        ],
    ) -> None: ...


class DatabaseConnection(Configuration, ABC):
    connection: Optional[Any]  # Replace with actual connection type
    _columns: List[Any]
    exit: Optional[Any]

    def __init__(
        self,
        config: Dict[str, Any],
        name: str = "",
        handlers: Optional[
            Dict[str, Union[Callable[..., Any], List[Callable[..., Any]]]]
        ] = None,
    ) -> None: ...

    def __enter__(self) -> DatabaseConnection: ...

    def __exit__(
        self,
        exc_type: Optional[type],
        exc_val: Optional[BaseException],
        exc_tb: Optional[Any],
    ) -> None: ...

    def execute_sp(
        self,
        procedure_name: str,
        params: Optional[Dict[str, Any]] = None,
        fetch: Optional[int] = None,
    ) -> Union[Dict[str, Any], List[Any]]: ...

    def execute(
        self,
        query: str,
        params: Union[
            Dict[str, str], Tuple[Tuple[str, str], ...], Tuple[str, str]
        ] = (),
        fetch: Optional[int] = None,
        offset: int = 0,
        limit: int = 0,
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]: ...

    def add_paging(
        self,
        query: str,
        offset: int = 0,
        limit: int = 0,
    ) -> str: ...

    def process_results(
        self,
        result: Union[Dict[str, Any], Result],
        query_type: QueryType = QueryType.QUERY,
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]: ...

    def fetch_results(
        self, fetch_size: int
    ) -> Union[Dict[str, Any], List[Any]]: ...

    def add_schema_prefix(
        self,
        query: str,
        query_type: Literal["procedure", "query"] = "query",
    ) -> str: ...

    def limit(self, limit: int) -> DatabaseConnection: ...

    def offset(self, offset: int) -> DatabaseConnection: ...

    def fetch(self, fetch_size: int) -> DatabaseConnection: ...

    def rollback(self) -> None: ...

    def commit(self) -> None: ...

    def close(self) -> None: ...

    def _set_activte_state(self) -> None: ...

    def _get_fetch_value(self, fetch_size: Optional[int]) -> Optional[int]: ...

    @abstractmethod
    def connect(self) -> Union[Connection, Session, scoped_session]: ...


class SqlAlchemyConnection(DatabaseConnection):
    _session_factory: Optional[Union[sessionmaker, scoped_session]]
    connection_type: Literal["scoped", "session", "direct"]

    def __init__(
        self,
        config: Dict[str, Any],
        name: str = "",
        connection_type: Literal["scoped", "session", "direct"] = "direct",
        session: Optional[Session] = None,
    ) -> None: ...

    @property
    def engine(self) -> Engine: ...

    def session(self) -> SqlAlchemyConnection: ...

    def connect(self) -> Union[Connection, Session, scoped_session]: ...

    def _create_session_factory(
        self,
    ) -> Union[sessionmaker, scoped_session]: ...

    def close(self) -> None: ...



"Database error messages."

from quik_db.utils.formatters import flatten_dict


db_errors = {
    "duplicate_value": """
        Duplicate value provided for {}.
    """,
    "handler_not_callable": """
        Handler must be callable.
    """,
    "invalid_connector": """
        Invalid connector specified. Only sqlalchemy, odbc, and dsn are
        allowed.
    """,
    "invalid_fetch": """
        Fetch value must be an integer. Use 0 for all rows.
    """,
    "invalid_fetch_size": """
        Fetch value must be an integer. Use 0 for all rows.
    """,
    "invalid_fetch_type": """
        Fetch value must be an integer. Use 0 for all rows.
    """,
    "invalid_session_type": """
        session_type must be in ('direct', 'scoped', 'session').
    """,
    "invalid_fetch_settings": """
        Fetch value is provided, but fetch is disabled.
    """,
    "key_error": """
        Please check your configuration file for a valid database entry.
    """,
    "missing_connector_with_url": """
        A valid connector is required when using a connection string.
    """,
    "missing_db_config": """
        Please check your configuration file for a valid database entry.
    """,
    "missing_dialect": """
        Please check your configuration file. A valid dialect or
        connection string is required.
    """,
    "missing_params": """
        Please check your configuration file. Missing connection
        information.
    """,
    "missing_type": """
        A connection string requires a type: ODBC, DSN, SQLAlchemy.
    """,
    "named_db_not_found": """
        The named database configuration was not found in the
        configuration file.
    """,
    "no_connection": """
        No connection was established.
    """,
    "no_connect_method": """
        You may see this error if you are attempting to use the
        DatabaseConnection class directly without subclassing. The
        DatabaseConnection class does not natively create connections
        itself, rather, it esbalishes a base configuration along with
        methods that can be used with a database.
    """,
    "no_database": """
        No database listed in the configuration file. Stored procedure
        may not work properly.
    """,
    "no_fetch_method": """
        The connection does not have a fetch method. You may see this
        errorif you have created a subclass and the connection method
        used does not have the method fetch. You must create a custom
        fetch method.
    """,
    "no_session": """
        A session was never created. Use connect() first.
    """,
    "no_result": """
        You have attempted to access the result object or a command that
        requires a result object, but the result object does not exist.
        Ensure you have executed a query statement before you fetch.
    """,
    "prefix_error": """
        Prefix requires a query, stored procedure, or function name.
    """,
    "not_implemented": """
        Method not implemented.
    """,
    "sqlite_no_sp": """
        SQLite does not support stored procedures or functions.
    """,
    "unsupported_dialect": """
        The database dialect is not supported. Please refer to the
        documentation for supported dialects.
    """,
    "unsupported_dialect_paging": """
        Unsupported database dialect for paging.
    """,
}

db_errors = flatten_dict(db_errors)
